"""Subversion interface."""

from __future__ import absolute_import

import re

import attr
import blinker
import svn.remote
import svn.common
from pillar import attrs_extra

task_logged = blinker.NamedSignal('task_logged')
shot_logged = blinker.NamedSignal('shot_logged')
marker_re = re.compile(r'\[(?P<type>[TS])(?P<node_id>[0-9a-zA-Z]+)\]')

signals = {
    'T': task_logged,
    'S': shot_logged,
}


def obtain(server_location):
    """Returns a Connection object for the given server location."""

    return svn.remote.RemoteClient(server_location)


@attr.s
class CommitLogObserver(object):
    svn_client = attr.ib(validator=attr.validators.instance_of(svn.remote.RemoteClient))
    last_seen_revision = attr.ib(default=0, validator=attr.validators.instance_of(int))
    _log = attrs_extra.log('%s.CommitLogObserver' % __name__)

    def fetch_and_observe(self):
        """Obtains task IDs from SVN logs."""

        self._log.debug('%s: fetch_and_observe()', self)

        try:
            svn_log = self.svn_client.log_default(revision_from=self.last_seen_revision + 1)
            for log_entry in svn_log:
                self._log.debug('- %r', log_entry)

                # assumption: revisions are always logged in strictly increasing order.
                self.last_seen_revision = log_entry.revision

                self._parse_log_entry(log_entry)
        except svn.common.SvnException:
            # The SVN library just raises a SvnCommandError when something goes wrong,
            # without any structured indication of the error. There isn't much else
            # we can do, except to log the error and return.
            self._log.exception('Error calling self.svn_client.log_default()')
            return

    def process_log(self, revision, commit_author, commit_message):
        """Obtains task IDs without accessing the SVN server directly."""

        self._log.debug('%s: process_log(%s, %s, ...)', self, revision, commit_author)
        for node_id, node_type, in self._find_ids(commit_message):
            signal = signals[node_type]
            signal.send(self, node_id=node_id, log_entry=commit_message, author=commit_author)

    def _find_ids(self, message):
        # Parse the commit log to see if there are any task/shot markers.
        lines = message.split('\n', 100)[:100]
        for line in lines:
            for match in marker_re.finditer(line):
                type = match.group('type')
                node_id = match.group('task_id')
                yield node_id, type

    def _parse_log_entry(self, log_entry):
        """Parses the commit log to see if there are any task markers."""

        # Log entries can be None.
        if not log_entry.msg:
            return

        # Parse the commit log to see if there are any task markers.
        lines = log_entry.msg.split('\n', 1)
        first_line = lines[0]
        for match in marker_re.finditer(first_line):
            task_id = match.group('task_id')

            # Send a Blinker signal for each observed task identifier.
            task_logged.send(self, task_id=task_id, log_entry=log_entry)
